# methods in cngi_io.image.image should be used going forward,
# methods in cngi_io.image.cngi_image_io are deprecated
from .image import load_image, make_empty_sky_image, read_image, write_image

__all__ = ["load_image", "make_empty_sky_image", "read_image", "write_image"]
